# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains generic utility functions for the library."""
import os
from json import JSONEncoder
from typing import Any, List, Optional, Type, Union

import numpy as np
from numpy.typing import NDArray

PathLike = Union[str, os.PathLike]


def _assert(
    condition: bool,
    error_message: Optional[str] = None,
    exception_class: Type[Exception] = Exception,
) -> None:
    """Check condition; if false, raise exception with the provided error message."""
    if not condition:
        raise exception_class(error_message)


def check_arg(condition: bool, message: str) -> None:
    """
    Check a condition over an argument (i.e. raises ValueError in case it is false).

    Args:
        condition: the condition to check
        message: the error message
    """
    _assert(condition, message, exception_class=ValueError)


def ensure_float(arg: Any) -> None:
    """Check that the argument is a float."""
    check_arg(isinstance(arg, (float, np.float64)), f"expected a float, got {arg}")


def positive_float(arg: float) -> float:
    """Check a float is positive."""
    ensure_float(arg)
    check_arg(arg >= 0, "provided float is not positive")
    return arg


def digitize_data(
    data: NDArray[np.float64], param_grid: List[NDArray[np.float64]]
) -> NDArray[np.float64]:
    """
    Return a discretized version of the input sorted_array.

    Args:
        data: the input array to be discretized
        param_grid: the discrete values allowed

    Returns:
        the discretized array
    """
    digitalized_data: NDArray[np.float64] = np.zeros(shape=data.shape)

    for i in range(data.shape[1]):
        digitalized_data[:, i] = get_closest(param_grid[i], data[:, i])

    return digitalized_data


def get_closest(sorted_array: NDArray, values: NDArray) -> NDArray:
    """Fast way to find the nearest element in 'sorted_array' for each element in 'values'.

    Solution taken from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array/46184652#46184652.

    Args:
        sorted_array: a sorted array
        values: an array of arbitrary values

    Returns:
        for each element in the 'values' array the closest element in 'sorted_array' is returned
    """
    # get insert positions
    idxs = np.searchsorted(sorted_array, values, side="left")

    # find indexes where previous index is closer
    prev_idx_is_less = (idxs == len(sorted_array)) | (
        np.fabs(values - sorted_array[np.maximum(idxs - 1, 0)])
        < np.fabs(values - sorted_array[np.minimum(idxs, len(sorted_array) - 1)])
    )
    idxs[prev_idx_is_less] -= 1

    return sorted_array[idxs]


def is_symmetric(a: NDArray[np.float64]) -> bool:
    """
    Check if a matrix is symmetric.

    Args:
        a: the matrix

    Returns:
        True if the matrix is symmetric
    """
    return a.shape[0] == a.shape[1] and (a == a.T).all()


class NumpyArrayEncoder(JSONEncoder):
    """
    Custom JSONEncoder for Numpy arrays.

    Solution from https://pynative.com/python-serialize-numpy-ndarray-into-json/.
    """

    def default(self, o: Any) -> Any:
        """Implement custom JSON serialization for NumPy NDArrays."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        return JSONEncoder.default(self, o)  # pragma: no cover


def get_random_seed(random_generator: np.random.Generator) -> int:
    """
    Get a random seed from a random generator.

    Sample an integer in the range [0, 2^32 - 1].

    Args:
        random_generator: the random generator to be used for sampling the random seed.

    Returns:
        the random seed.
    """
    return random_generator.integers(2**32 - 1)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the Gaussian process-based sampling."""
import random
import warnings
from enum import Enum
from typing import Optional, Tuple, cast

import GPy
import numpy as np
from GPy.models import GPRegression
from numpy.typing import NDArray
from scipy.special import erfc  # pylint: disable=no-name-in-module

from black_it.samplers.surrogate import MLSurrogateSampler


class _AcquisitionTypes(Enum):
    """Enumeration of allowed acquisition types."""

    MEAN = "mean"
    EI = "expected_improvement"

    def __str__(self) -> str:
        """Get the string representation."""
        return self.value


class GaussianProcessSampler(MLSurrogateSampler):
    """
    This class implements the Gaussian process-based sampler.

    In particular, the sampling is based on a Gaussian Process interpolation of the loss function.

    Note: this class is a wrapper of the GPRegression model of the GPy package.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_deduplication_passes: int = 5,
        candidate_pool_size: Optional[int] = None,
        max_iters: int = 1000,
        optimize_restarts: int = 5,
        acquisition: str = "expected_improvement",
    ):
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: the maximum number of deduplication passes that are made
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            max_iters: maximum number of iteration in the optimization of the GP hyperparameters
            optimize_restarts: number of independent random trials of the optimization of the GP hyperparameters
            acquisition: type of acquisition function, it can be 'expected_improvement' of simply 'mean'
        """
        self._validate_acquisition(acquisition)

        super().__init__(
            batch_size, random_state, max_deduplication_passes, candidate_pool_size
        )
        self.max_iters = max_iters
        self.optimize_restarts = optimize_restarts
        self.acquisition = acquisition
        self._gpmodel: Optional[GPRegression] = None

    @staticmethod
    def _validate_acquisition(acquisition: str) -> None:
        """
        Check that the required acquisition is among the supported ones.

        Args:
            acquisition: the acquisition provided as input of the constructor.

        Raises
            ValueError: if the provided acquisition type is not among the allowed ones.
        """
        try:
            _AcquisitionTypes(acquisition)
        except ValueError as e:
            raise ValueError(
                "expected one of the following acquisition types: "
                f"[{' '.join(map(str, _AcquisitionTypes))}], "
                f"got {acquisition}"
            ) from e

    def fit(self, X: NDArray[np.float64], y: NDArray[np.float64]) -> None:
        """Fit a gaussian process surrogate model."""
        y = np.atleast_2d(y).T

        if X.shape[0] > 500:
            warnings.warn(
                "Standard GP evaluations can be expensive for large datasets, consider implementing a sparse GP",
                RuntimeWarning,
            )

        # initialize GP class from GPy with a Matern kernel by default
        dims = X.shape[1]
        kern = GPy.kern.Matern52(dims, variance=1.0, ARD=False)
        noise_var = y.var() * 0.01

        self._gpmodel = GPRegression(
            X, y, kernel=kern, noise_var=noise_var, mean_function=None
        )

        # Make sure we do not get ridiculously small residual noise variance
        self._gpmodel.Gaussian_noise.constrain_bounded(
            1e-9, 1e6, warning=False
        )  # constrain_positive(warning=False)

        # we need to set the seed globally for GPy optimisations
        # to give reproducible results
        np.random.seed(self._get_random_seed())
        random.seed(self._get_random_seed())
        if self.max_iters > 0:
            # --- update the model maximizing the marginal likelihood.
            if self.optimize_restarts == 1:
                self._gpmodel.optimize(
                    optimizer="bfgs",
                    max_iters=self.max_iters,
                    messages=False,
                    ipython_notebook=False,
                )
            else:
                self._gpmodel.optimize_restarts(
                    num_restarts=self.optimize_restarts,
                    optimizer="bfgs",
                    max_iters=self.max_iters,
                    verbose=False,
                )

    def predict(self, X: NDArray[np.float64]) -> NDArray[np.float64]:
        """Predict using a gaussian process surrogate model."""
        # predict mean or expected improvement on the full sample set
        if self.acquisition == _AcquisitionTypes.EI.value:
            # minus sign needed for subsequent sorting
            candidates_score = -self._predict_EI(X)[:, 0]
        else:  # acquisition is "mean"
            candidates_score = self._predict_mean_std(X)[0][:, 0]

        return candidates_score

    def _predict_mean_std(
        self, X: NDArray[np.float64]
    ) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Predict mean and standard deviation of a fitted GP.

        Args:
            X: the points on which the predictions should be performed

        Returns:
            The pair (mean, std).
        """
        gpmodel = cast(GPRegression, self._gpmodel)
        X = X[None, :] if X.ndim == 1 else X
        m, v = gpmodel.predict(X, full_cov=False, include_likelihood=True)
        v = np.clip(v, 1e-10, np.inf)
        return m, np.sqrt(v)

    def _get_fmin(self) -> float:
        """Return the location where the posterior mean is takes its minimal value."""
        gpmodel = cast(GPRegression, self._gpmodel)
        return gpmodel.predict(gpmodel.X)[0].min()

    def _predict_EI(
        self, X: NDArray[np.float64], jitter: float = 0.1
    ) -> NDArray[np.float64]:
        """
        Compute the Expected Improvement per unit of cost.

        Args:
            X:  the points on which the predictions should be performed
            jitter: positive value to make the acquisition more explorative.

        Returns:
            the expected improvement.
        """
        m, s = self._predict_mean_std(X)

        fmin = self._get_fmin()

        phi, Phi, u = self.get_quantiles(jitter, fmin, m, s)

        f_acqu = s * (u * Phi + phi)

        return f_acqu

    @staticmethod
    def get_quantiles(
        acquisition_par: float,
        fmin: float,
        m: NDArray[np.float64],
        s: NDArray[np.float64],
    ) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """
        Quantiles of the Gaussian distribution useful to determine the acquisition function values.

        Args:
            acquisition_par: parameter of the acquisition function
            fmin: current minimum.
            m: vector of means.
            s: vector of standard deviations.

        Returns:
            the quantiles.
        """
        # remove values of variance that are too small
        s[s < 1e-10] = 1e-10

        u: NDArray[np.float64] = (fmin - m - acquisition_par) / s
        phi: NDArray[np.float64] = np.exp(-0.5 * u**2) / np.sqrt(2 * np.pi)
        Phi: NDArray[np.float64] = 0.5 * erfc(-u / np.sqrt(2))

        return phi, Phi, u



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the Fast Fourier Transform loss."""
from typing import Callable, List, Optional

import numpy as np
from numpy.typing import NDArray

from black_it.loss_functions.base import BaseLoss
from black_it.utils.base import _assert

FrequencyFilter = Callable[[NDArray[np.complex128], float], NDArray[np.complex128]]
"""A filter that receives the signal in the frequency domain and returns its
filtered version. Used by FourierLoss constructor.

In this version, the filter supports a single parameter: no multi-band filtering
is supported yet.
"""


def ideal_low_pass_filter(
    signal_frequencies: NDArray[np.complex128],
    f: float,
) -> NDArray[np.complex128]:
    """Ideal low-pass filter.

    Args:
        signal_frequencies: the input signal transformed in the frequency
            domain.
        f: the fraction of frequencies to keep unchanged, for f=1 the filter is
            just the identity.

    Returns:
        the filtered frequencies
    """
    # number of low-frequency component to keep
    n = int(np.round(f * signal_frequencies.shape[0]))

    # ideal low-pass filter
    mask = np.zeros(signal_frequencies.shape[0])
    mask[:n] = 1.0
    filtered_frequencies = signal_frequencies * mask

    return filtered_frequencies


def gaussian_low_pass_filter(
    signal_frequencies: NDArray[np.complex128],
    f: float,
) -> NDArray[np.complex128]:
    """Gaussian low-pass filter.

    Args:
        signal_frequencies: the input signal transformed in the frequency
            domain.
        f: the fraction of frequencies to keep, this will determine the length
            scale of the Gaussian filter

    Returns:
        the filtered frequencies
    """
    # number of low-frequency component to keep
    sigma = np.round(f * signal_frequencies.shape[0])

    # gaussian low-pass filter
    mask = np.exp(-np.arange(signal_frequencies.shape[0]) ** 2 / (2 * sigma**2))
    filtered_frequencies = signal_frequencies * mask

    return filtered_frequencies


class FourierLoss(BaseLoss):
    """Class for the Fourier loss."""

    def __init__(
        self,
        frequency_filter: FrequencyFilter = gaussian_low_pass_filter,
        f: float = 0.8,
        coordinate_weights: Optional[NDArray] = None,
        coordinate_filters: Optional[List[Optional[Callable]]] = None,
    ) -> None:
        """Loss computed using a distance in the Fourier space of the time series.

        This loss is equivalent to the Euclidean loss computed on the time
        series after a Fourier-filter.
        The parameter f controls the fraction of frequencies that are kept in
        the Fourier series.

        Args:
            frequency_filter: the function used to filter the fourier
                frequencies before the distance is computed.
            f: fraction of fourier components to keep when computing the
                distance between the time series. This parameter will be passed
                to frequency_filter.
            coordinate_weights: relative weights of the losses computed over
                different time series coordinates.
            coordinate_filters: filters/transformations to be applied to each simulated series before
                the loss computation.
        """
        _assert(
            0.0 < f <= 1.0,
            "'f' must be in the interval (0.0, 1.0]",
        )
        self.frequency_filter = frequency_filter
        self.f = f
        super().__init__(coordinate_weights, coordinate_filters)

    def compute_loss_1d(
        self, sim_data_ensemble: NDArray[np.float64], real_data: NDArray[np.float64]
    ) -> float:
        """Compute Euclidean distance between the Fourier transform of the two time series.

        Args:
            sim_data_ensemble: the first operand
            real_data: the second operand

        Returns:
            The computed loss over the specific coordinate.
        """
        f_real_data = np.fft.rfft(real_data, axis=0)
        N = f_real_data.shape[0]
        f_real_data = self.frequency_filter(f_real_data, self.f)
        # computer mean fft transform of simulated ensemble
        f_sim_data = []
        for s in sim_data_ensemble:
            f_sim_data_ = np.fft.rfft(s, axis=0)
            f_sim_data_ = self.frequency_filter(f_sim_data_, self.f)
            f_sim_data.append(f_sim_data_)

        f_sim_data = np.array(f_sim_data).mean(0)

        loss_1d = np.sqrt(np.sum((abs(f_sim_data - f_real_data)) ** 2) / N)

        return loss_1d



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains utility functions to deal with time series."""

from typing import Tuple

import numpy as np
import scipy.sparse as sps
import statsmodels.api as sm
from numpy.typing import NDArray
from scipy.stats import kurtosis, skew


def get_mom_ts(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments
    """
    moments = np.array([get_mom_ts_1d(ts_d) for ts_d in time_series.T])
    return moments.T


def get_mom_ts_1d(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments from a time series
    """
    # array of all moments (to be filled)
    avg_vec_mom = np.zeros(18)

    # first 4 moments and auto-correlations of the time series
    avg_vec_mom[0] = np.mean(time_series)
    avg_vec_mom[1] = np.std(time_series)
    s, k = skew(time_series), kurtosis(time_series)  # pylint: disable=invalid-name
    avg_vec_mom[2] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[3] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_acf = sm.tsa.acf(time_series, nlags=5, fft=False)
    avg_vec_mom[4] = ts_acf[1]
    avg_vec_mom[5] = ts_acf[2]
    avg_vec_mom[6] = ts_acf[3]
    avg_vec_mom[7] = ts_acf[4]
    avg_vec_mom[8] = ts_acf[5]

    # first 4 moments and auto-correlations of the absolute differences of the time series
    abs_diff = np.absolute(np.diff(time_series))

    avg_vec_mom[9] = np.mean(abs_diff)
    avg_vec_mom[10] = np.std(abs_diff)
    s, k = skew(abs_diff), kurtosis(abs_diff)  # pylint: disable=invalid-name
    avg_vec_mom[11] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[12] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_diff_acf = sm.tsa.acf(abs_diff, nlags=5, fft=False)
    avg_vec_mom[13] = ts_diff_acf[1]
    avg_vec_mom[14] = ts_diff_acf[2]
    avg_vec_mom[15] = ts_diff_acf[3]
    avg_vec_mom[16] = ts_diff_acf[4]
    avg_vec_mom[17] = ts_diff_acf[5]

    np.nan_to_num(avg_vec_mom, False)

    return avg_vec_mom


def hp_filter(
    time_series: NDArray[np.float64], lamb: float = 1600
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """
    Apply the HP filter to a time series.

    Args:
        time_series: the one dimensional series to be filtered
        lamb: the Hodrick-Prescott smoothing parameter, a value of 1600 is suggested for quarterly data

    Returns:
        the cycle and the trend of the HP decomposition of the time series

    References:
        Hodrick, R.J, and E. C. Prescott. 1980. "Postwar U.S. Business Cycles: An Empirical Investigation."
        Carnegie Mellon University discussion paper no. 451.
    """
    nobs = len(time_series)
    I = sps.eye(nobs, nobs)  # noqa:E741
    offsets = np.array([0, 1, 2])
    data = np.repeat([[1.0], [-2.0], [1.0]], nobs, axis=1)
    K = sps.dia_matrix((data, offsets), shape=(nobs - 2, nobs))

    trend = sps.linalg.spsolve(I + lamb * K.T.dot(K), time_series, use_umfpack=True)
    cycle = time_series - trend

    return cycle, trend


def hp_cycle_lamb1600_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Get the cycle part of the HP filter to a time series using lamb = 1600.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the cycle part of the HP decomposition of the time series
    """
    return hp_filter(time_series, lamb=1600)[0]


def log_and_hp_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Take the log of the values and remove the HP trend from them.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the filtered time series
    """
    transformed_series = (
        np.log(time_series) - hp_filter(np.log(time_series), lamb=1600)[1]
    )
    return transformed_series


def diff_log_demean_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Take the difference of log of the values and remove the mean from them.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the filtered time series
    """
    log = np.log(time_series)
    diff_log = np.diff(log, prepend=log[0])
    transformed_series = diff_log - np.mean(diff_log)
    return transformed_series



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the definition of the search space abstractions."""

from typing import List, Union

import numpy as np
from numpy.typing import NDArray


class SearchSpace:  # pylint: disable=too-few-public-methods
    """A class that contains information on the search grid of explorable parameters."""

    def __init__(
        self,
        parameters_bounds: Union[NDArray[np.float64], List[List[float]]],
        parameters_precision: Union[NDArray[np.float64], List[float]],
        verbose: bool,
    ):
        """
        Initialize the SearchSpace object.

        The values of parameters_bounds and parameters_precision parameters have
        to satisfy the following constraints, otherwise an exception (subclass
        of SearchSpaceError) is raised:

        - parameters_bounds must be a two-elements array/list (or
          BoundsNotOfSizeTwoError is raised)
        - the two sub-arrays/lists of parameter_bounds must have the same length (or
          BoundsOfDifferentLengthError is raised)
        - parameters_precision array/list must have the same number of elements than
          the two sub-arrays/lists in parameters_bounds (or BadPrecisionLengthError is
          raised)
        - lower bounds and upper bounds cannot have the same value (or
          SameLowerAndUpperBoundError is raised)
        - every lower bound must be lower than the corresponding upper bound (or
          LowerBoundGreaterThanUpperBoundError)
        - 0 is an invalid value for a precision (or PrecisionZeroError is
          raised)
        - any given parameter precision has to be strictly lower than the
          allowed parameter span (or PrecisionGreaterThanBoundsRangeError is
          raised)

        Args:
            parameters_bounds: lower and upper bounds of the parameters.
            parameters_precision: resolution of the grid of parameters.
            verbose: whether to print or not the information on the search space.
        """
        SearchSpace._check_bounds(parameters_bounds, parameters_precision)

        # The bounds we were given are well formed. Save them.
        self._parameters_bounds = np.array(parameters_bounds)
        self._parameters_precision = np.array(parameters_precision)

        # Initialize search grid
        self._param_grid: List[NDArray[np.float64]] = []
        self._space_size = 1
        for i in range(self.dims):
            new_col = np.arange(
                parameters_bounds[0][i],
                parameters_bounds[1][i] + 0.0000001,
                parameters_precision[i],
                dtype=np.float64,
            )
            self._param_grid.append(new_col)
            self._space_size *= len(new_col)

        if verbose:
            print("\n***")
            print(f"Number of free params:       {self.dims}.")
            print(f"Explorable param space size: {self.space_size}.")
            print("***\n")

    @staticmethod
    def _check_bounds(
        parameters_bounds: Union[NDArray[np.float64], List[List[float]]],
        parameters_precision: Union[NDArray[np.float64], List[float]],
    ) -> None:
        """
        Ensure parameter_bounds and parameter_precision have acceptable values.

        This is an helper function for SearchSpace.__init__().

        Args:
            parameters_bounds: lower and upper bounds of the parameters
            parameters_precision: resolution of the grid of parameters
        """
        # ensure parameters_bounds is a two-elements array
        if len(parameters_bounds) != 2:
            raise BoundsNotOfSizeTwoError(len(parameters_bounds))

        # ensure the two sub-arrays of parameter_bounds (which are the min and
        # max bounds for each parameter) have the same length
        if len(parameters_bounds[0]) != len(parameters_bounds[1]):
            raise BoundsOfDifferentLengthError(
                len(parameters_bounds[0]),
                len(parameters_bounds[1]),
            )

        # ensure parameters_precision array has as many elements as either one
        # of parameters_bounds sub-array
        if len(parameters_precision) != len(parameters_bounds[0]):
            raise BadPrecisionLengthError(
                len(parameters_precision),
                len(parameters_bounds[0]),
            )

        for i, (lower_bound, upper_bound, precision) in enumerate(
            zip(parameters_bounds[0], parameters_bounds[1], parameters_precision)
        ):
            # ensure lower bounds and upper bounds do not have the same value
            if lower_bound == upper_bound:
                raise SameLowerAndUpperBoundError(i, lower_bound)

            # ensure each lower bound is lower than the corresponding upper one
            if lower_bound > upper_bound:
                raise LowerBoundGreaterThanUpperBoundError(i, lower_bound, upper_bound)

            # a precision of 0 is not meaningful
            if precision == 0:
                raise PrecisionZeroError(i)

            # any given parameter precision has to be strictly lower than the
            # allowed parameter span
            if precision > (upper_bound - lower_bound):
                raise PrecisionGreaterThanBoundsRangeError(
                    i, lower_bound, upper_bound, precision
                )

    @property
    def param_grid(self) -> List[NDArray[np.float64]]:
        """Discretized parameter space containing all possible candidates for calibration."""
        return self._param_grid

    @property
    def parameters_bounds(self) -> NDArray[np.float64]:
        """Two dimensional array containing lower and upper bounds for each parameter."""
        return self._parameters_bounds

    @property
    def parameters_precision(self) -> NDArray[np.float64]:
        """One dimensional array containing the precisions for each parameter."""
        return self._parameters_precision

    @property
    def dims(self) -> int:
        """Return the number of model parameters configured for calibration."""
        return len(self._parameters_precision)

    @property
    def space_size(self) -> int:
        """Cardinality of the potential parameter space to be searched."""
        return self._space_size


class SearchSpaceError(ValueError):
    """Base class for the exceptions raised by SearchSpace when its construction fails.

    If you need to distinguish a specific error, please do not rely on parsing
    the error message, because it is considered unstable: catch instead the
    specific exception type you want to handle, and use the additional fields
    each subtype exposes (for example: BadPrecisionLengthError.bounds_length)
    """


class BoundsNotOfSizeTwoError(SearchSpaceError):
    """Raised when bounds are not a two-dimensional array.

    Attributes:
        count_bounds_subarrays (int): wrong number of subarrays the
            parameter_bounds array was made of. It will be different than 2.
    """

    def __init__(self, count_bounds_subarrays: int) -> None:  # noqa: D107
        super().__init__(
            f"parameters_bounds must be a two dimensional array. This one has "
            f"size {count_bounds_subarrays}."
        )
        self.count_bounds_subarrays = count_bounds_subarrays


class BoundsOfDifferentLengthError(SearchSpaceError):
    """Raised when the lower and upper bounds do not have the same number of elements.

    Attributes:
        lower_bounds_length (int): number of elements in the subarray 0. It will
            be different than upper_bounds_length
        upper_bounds_length (int): number of elements in the subarray 1. It will
            be different than lower_bounds_length
    """

    def __init__(  # noqa: D107
        self, lower_bounds_length: int, upper_bounds_length: int
    ) -> None:
        super().__init__(
            f"parameters_bounds subarrays must be of the same length. Lower "
            f"bounds length: {lower_bounds_length}, upper bounds length: "
            f"{upper_bounds_length}."
        )
        self.lower_bounds_length = lower_bounds_length
        self.upper_bounds_length = upper_bounds_length


class BadPrecisionLengthError(SearchSpaceError):
    """Raised when the parameters_precision array has a different length than the bounds'.

    Attributes:
        precisions_length (int): number of elements of the precision subarray
        bounds_length (int): number of elements of the two bounds subbarrays
    """

    def __init__(  # noqa: D107
        self, precisions_length: int, bounds_length: int
    ) -> None:
        super().__init__(
            f"parameters_precision array has {precisions_length} elements. Its "
            f"length, instead, has to be {bounds_length}, the same as the bounds'."
        )
        self.precisions_length = precisions_length
        self.bounds_length = bounds_length


class SameLowerAndUpperBoundError(SearchSpaceError):
    """Raised when the lower and the upper bound of a parameter have the same value.

    Attributes:
        param_index (int): 0-based index of the parameter presenting the error
        bound_value (float): common value of the parameter bound
    """

    def __init__(self, param_index: int, bound_value: float) -> None:  # noqa: D107
        super().__init__(
            f"Parameter {param_index}'s lower and upper bounds have the same "
            f"value ({bound_value}). This calibrator cannot handle that. "
            f"Please redefine externally your model in order to hardcode "
            f"parameter {param_index} to {bound_value}."
        )
        self.param_index = param_index
        self.bound_value = bound_value


class LowerBoundGreaterThanUpperBoundError(SearchSpaceError):
    """Raised when the lower bound of a parameter is greater than its upper bound.

    Attributes:
        param_index (int): 0-based index of the parameter presenting the error
        lower_bound (float): lower bound. It will be higher than upper bound
        upper_bound (float): upper bound. It will be lower than lower bound
    """

    def __init__(  # noqa: D107
        self, param_index: int, lower_bound: float, upper_bound: float
    ) -> None:
        super().__init__(
            f"Parameter {param_index}'s lower bound ({lower_bound}) must be "
            f"lower than its upper bound ({upper_bound})."
        )
        self.param_index = param_index
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound


class PrecisionZeroError(SearchSpaceError):
    """Raised when a parameter precision is set to 0.

    Attributes:
        param_index (int): 0-based index of the parameter presenting the error
    """

    def __init__(self, param_index: int) -> None:  # noqa: D107
        super().__init__(f"Parameter {param_index}'s precision cannot be zero.")
        self.param_index = param_index


class PrecisionGreaterThanBoundsRangeError(SearchSpaceError):
    """Raised when the precision step is greater than the parameter range.

    Attributes:
        param_index (int): 0-based index of the parameter presenting the error
        lower_bound (float): lower bound
        upper_bound (float): upper bound
        precision (float): the illegal precision. It will be greater than
            (upper_bound - lower_bound)
    """

    def __init__(  # noqa: D107
        self, param_index: int, lower_bound: float, upper_bound: float, precision: float
    ) -> None:
        super().__init__(
            f"Parameter {param_index}'s allowed range is [{lower_bound}, "
            f"{upper_bound}]. The requested precision is {precision}, but this "
            f"number cannot be greater than the bounds range "
            f"({upper_bound}-{lower_bound}={upper_bound-lower_bound})."
        )
        self.param_index = param_index
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.precision = precision



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains serialization and deserialization of calibration state with SQLite."""
import gzip
import io
import json
import pickle  # nosec B403
import sqlite3
from pathlib import Path
from typing import Mapping, Optional, Sequence, Tuple

import numpy as np
from numpy.typing import NDArray

from black_it.loss_functions.base import BaseLoss
from black_it.samplers.base import BaseSampler
from black_it.utils.base import PathLike

SCHEMA_VERSION = 3
"""
Encodes the version of a checkpoint, preventing to load incompatible versions.

Each time the structure of the SQL changes, this number has to be incremented.
A calibrator whose schema version is X will not be able to load checkpoints
saved with schema version Y, unless conversion mechanisms are in place.
"""

SQL_LOAD_QUERY = """
    SELECT
        parameters_bounds,
        parameters_precision,
        real_data,
        ensemble_size,
        N,
        D,
        convergence_precision,
        verbose,
        saving_folder,
        initial_random_seed,
        random_generator_state_json,
        model_name,
        samplers_pickled,
        loss_function_pickled,
        current_batch_index,
        params_samp,
        losses_samp,
        series_samp,
        batch_num_samp,
        method_samp
    FROM checkpoint
"""

SQL_SAVE_QUERY = """
    INSERT INTO checkpoint (
        parameters_bounds,
        parameters_precision,
        real_data,
        ensemble_size,
        N,
        D,
        convergence_precision,
        verbose,
        saving_folder,
        initial_random_seed,
        random_generator_state_json,
        model_name,
        samplers_pickled,
        loss_function_pickled,
        current_batch_index,
        params_samp,
        losses_samp,
        series_samp,
        batch_num_samp,
        method_samp
    ) VALUES (
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?,
        ?
    )
"""

SQL_LOAD_USER_VERSION = """
    PRAGMA user_version
"""

SQL_SAVE_USER_VERSION = f"""
    PRAGMA user_version={SCHEMA_VERSION}
"""

SQL_DDL = """
    CREATE TABLE IF NOT EXISTS checkpoint (
        -- initialization parameters
        parameters_bounds            NDARRAY,
        parameters_precision         NDARRAY,
        real_data                    NDARRAY,
        ensemble_size                INTEGER,
        N                            INTEGER,
        D                            INTEGER,
        convergence_precision        DOUBLE,
        verbose                      INTEGER,
        saving_folder                TEXT,
        initial_random_seed          INTEGER,
        random_generator_state_json  TEXT,
        model_name                   TEXT,
        samplers_pickled             BLOB,
        loss_function_pickled        BLOB,
        -- arrays resulting from calibration
        current_batch_index          INTEGER,
        params_samp                  NDARRAY,
        losses_samp                  NDARRAY,
        series_samp                  GZ_NDARRAY,
        batch_num_samp               NDARRAY,
        method_samp                  NDARRAY
    );

    DELETE FROM checkpoint;
"""


class gz_ndarray(NDArray):
    """Convenience class to tell sqlite which ndarrays are to be compressed.

    If a user wants to compress/decompress the serialized version of an NDArray
    when saving it to the database, he can cast it to gz_ndarray without copying
    any memory:

        some_ndarray: NDArray = [1.0, 2.0]
        please_compressme = some_ndarray.view(gz_ndarray)

    please_compressme will then be a view on the same memory of some_ndarray,
    but its type will be gz_ndarray. An sqlite3 adapter will then be able to
    detect at INSERT time that the variable is of type gz_ndarray and compress
    it.
    """


def load_calibrator_state(  # pylint: disable=too-many-locals
    checkpoint_path: PathLike,
) -> Tuple:
    """
    Load the calibration state.

    Checks that the schema version stored in checkpoint_path has the same value
    of SCHEMA_VERSION. Currently there are no conversion mechanisms in place, so
    loading will only succeed if the schema versions are exactly the same.

    Args:
        checkpoint_path: the path to the checkpoint directory.

    Returns:
        a tuple with all the data that make a state.
    """
    checkpoint_path = Path(checkpoint_path)
    connection = sqlite3.connect(
        checkpoint_path / "checkpoint.sqlite", detect_types=sqlite3.PARSE_DECLTYPES
    )
    try:
        cursor = connection.cursor()

        checkpoint_schema_version: int = cursor.execute(
            SQL_LOAD_USER_VERSION
        ).fetchone()[0]
        if checkpoint_schema_version != SCHEMA_VERSION:
            raise Exception(
                f"The checkpoint you want to load has been generated with another version of the code:\n"
                f"\tCheckpoint schema version:          {checkpoint_schema_version}"
                f"\tSchema version of the current code: {SCHEMA_VERSION}"
            )

        (
            parameters_bounds,
            parameters_precision,
            real_data,
            ensemble_size,
            N,
            D,
            convergence_precision,
            verbose,
            saving_file,
            initial_random_seed,
            random_generator_state_json,
            model_name,
            samplers_pickled,
            loss_function_pickled,
            current_batch_index,
            params_samp,
            losses_samp,
            series_samp,
            batch_num_samp,
            method_samp,
        ) = cursor.execute(SQL_LOAD_QUERY).fetchone()

        samplers = pickle.loads(samplers_pickled)  # nosec B301
        loss_function = pickle.loads(loss_function_pickled)  # nosec B301

        return (
            parameters_bounds,
            parameters_precision,
            real_data,
            ensemble_size,
            N,
            D,
            convergence_precision,
            verbose,
            saving_file,
            initial_random_seed,
            json.loads(random_generator_state_json),
            model_name,
            samplers,
            loss_function,
            current_batch_index,
            params_samp,
            losses_samp,
            series_samp,
            batch_num_samp,
            method_samp,
        )

    except BaseException as err:  # pylint: disable=broad-except
        raise err from err
    finally:
        connection.close()


def save_calibrator_state(  # pylint: disable=too-many-arguments,too-many-locals
    checkpoint_path: PathLike,
    parameters_bounds: NDArray[np.float64],
    parameters_precision: NDArray[np.float64],
    real_data: NDArray[np.float64],
    ensemble_size: int,
    N: int,
    D: int,
    convergence_precision: Optional[float],
    verbose: bool,
    saving_file: Optional[str],
    initial_random_seed: Optional[int],
    random_generator_state: Mapping,
    model_name: str,
    samplers: Sequence[BaseSampler],
    loss_function: BaseLoss,
    current_batch_index: int,
    params_samp: NDArray[np.float64],
    losses_samp: NDArray[np.float64],
    series_samp: NDArray[np.float64],
    batch_num_samp: NDArray[np.int64],
    method_samp: NDArray[np.int64],
) -> None:
    """
    Save a calibration state.

    Args:
        checkpoint_path: path to the checkpoint
        parameters_bounds: the parameters bounds
        parameters_precision: the parameters precision
        real_data: the real data
        ensemble_size: the ensemble size
        N: the number of samples
        D: the number of dimensions
        convergence_precision: the convergence precision
        verbose: the verbosity mode
        saving_file: the saving file
        initial_random_seed: initial random seed for the calibration task
        random_generator_state: the state of the random generator
        model_name: the model name
        samplers: the ordered list of samplers to use in the calibration
        loss_function: the loss function
        current_batch_index: the current batch index
        params_samp: the sampled parameters
        losses_samp: the sampled losses
        series_samp: the sampled series
        batch_num_samp: the sampling batch number
        method_samp: the sampling method
    """
    checkpoint_path = Path(checkpoint_path)
    # create directory if needed
    if not checkpoint_path.exists():
        checkpoint_path.mkdir(parents=True)

    # pickle some files
    samplers_pickled = pickle.dumps(samplers)
    loss_function_pickled = pickle.dumps(loss_function)

    connection = sqlite3.connect(
        checkpoint_path / "checkpoint.sqlite", detect_types=sqlite3.PARSE_DECLTYPES
    )

    try:
        cursor = connection.cursor()
        cursor.execute(SQL_SAVE_USER_VERSION)
        cursor.executescript(SQL_DDL)

        cursor.execute(
            SQL_SAVE_QUERY,
            (
                parameters_bounds,
                parameters_precision,
                real_data,
                ensemble_size,
                N,
                D,
                convergence_precision,
                verbose,
                saving_file,
                initial_random_seed,
                json.dumps(random_generator_state),
                model_name,
                samplers_pickled,
                loss_function_pickled,
                current_batch_index,
                params_samp,
                losses_samp,
                series_samp.view(gz_ndarray),  # require compression on series_samp
                batch_num_samp,
                method_samp,
            ),
        )

        connection.commit()

    except BaseException as err:  # pylint: disable=broad-except
        connection.rollback()
        raise err from err
    finally:
        connection.close()


def npndarray_to_sqlite_binary(numpy_array: NDArray) -> sqlite3.Binary:
    """
    Serialize an NumPy NDArray to SQLite binary.

    Taken from: http://stackoverflow.com/a/31312102/190597 (SoulNibbler)

    Args:
        numpy_array: the numpy array to be serialized.

    Returns:
        the serialized NumPy array.
    """
    out: io.BytesIO = io.BytesIO()
    np.save(out, numpy_array)
    out.seek(0)
    return sqlite3.Binary(out.read())


def sqlite_binary_to_npndarray(sqlite_binary: bytes) -> NDArray:
    """
    Deserialize a NumPy NDArray from SQLite binary.

    Taken from: http://stackoverflow.com/a/31312102/190597 (SoulNibbler)

    Args:
        sqlite_binary: the SQLite binary version of the array.

    Returns:
        the deserialized NumPy NDArray.
    """
    out: io.BytesIO = io.BytesIO(sqlite_binary)
    out.seek(0)
    out = io.BytesIO(out.read())
    return np.load(out)


def gz_ndarray_to_gzipped_sqlite_binary(numpy_array: gz_ndarray) -> sqlite3.Binary:
    """
    Serialize a NumPy NDArray to a gzipped SQLite binary.

    The custom gz_ndarray type is a convenience type used to trigger the
    sqlite3 adapter mechanism. If a user coerces a NDArray to gz_ndarray, it
    will be automatically gzipped after serialization.

    Args:
        numpy_array: the numpy array to be serialized.

    Returns:
        the compressed serialized NumPy array.
    """
    out: io.BytesIO = io.BytesIO()

    with gzip.GzipFile(fileobj=out, mode="wb") as gzf:
        np.save(gzf, numpy_array)
    out.seek(0)
    return sqlite3.Binary(out.read())


def gzipped_sqlite_binary_to_npndarray(sqlite_binary: bytes) -> NDArray:
    """
    Deserialize a NumPy NDArray from a gzipped SQLite binary.

    Args:
        sqlite_binary: the gzipped SQLite binary version of the array.

    Returns:
        the deserialized NumPy NDArray.
    """
    buf: io.BytesIO = io.BytesIO(sqlite_binary)
    buf.seek(0)
    buf = io.BytesIO(buf.read())
    with gzip.GzipFile(fileobj=buf, mode="rb") as gzf:
        return np.load(gzf)


# Converts an np.array to a BLOB when INSERTing
sqlite3.register_adapter(np.ndarray, npndarray_to_sqlite_binary)

# Converts a BLOB to an np.ndarray when SELECTing
sqlite3.register_converter("NDARRAY", sqlite_binary_to_npndarray)

# Converts the custom type gz_ndarray to a gzipped BLOB when INSERTing
sqlite3.register_adapter(gz_ndarray, gz_ndarray_to_gzipped_sqlite_binary)

# Converts a gzipped BLOB to an np.ndarray when SELECTing
sqlite3.register_converter("GZ_NDARRAY", gzipped_sqlite_binary_to_npndarray)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains utilities for plotting results."""
import os
import pickle  # nosec B403
from typing import Collection, Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from ipywidgets import fixed, interact

from black_it.calibrator import Calibrator


def _get_samplers_id_table(saving_folder: Union[str, os.PathLike]) -> Dict[str, int]:
    """Get the id table of the samplers from the checkpoint.

    Args:
        saving_folder: the folder where the calibration results were saved

    Returns:
        the id table of the samplers
    """
    with open(os.path.join(saving_folder, "samplers_pickled.pickle"), "rb") as f:
        method_list = pickle.load(f)  # nosec B301

    samplers_id_table = (
        Calibrator._construct_samplers_id_table(  # pylint: disable=protected-access
            method_list
        )
    )

    return samplers_id_table


def _get_samplers_names(
    saving_folder: Union[str, os.PathLike], ids: List[int]
) -> List[str]:
    """Get the names of the samplers from their ids and from the checkpoint of the calibration.

    Args:
        saving_folder: the folder where the calibration results were saved
        ids: the ids of the samplers

    Returns:
        the names of the samplers
    """
    samplers_id_table = _get_samplers_id_table(saving_folder)

    inv_samplers_id_table = {v: k for k, v in samplers_id_table.items()}

    sampler_names = []
    for sampler_id in ids:
        sampler_names.append(inv_samplers_id_table[sampler_id])

    return sampler_names


def plot_convergence(saving_folder: Union[str, os.PathLike]) -> None:
    """Plot the loss values sampled by the various methods as a function of the batch number.

    Args:
        saving_folder: the folder where the calibration results were saved

    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))

    losses_cummin = df.groupby("batch_num_samp").min()["losses_samp"].cummin()

    plt.figure()
    g = sns.lineplot(
        data=df, x="batch_num_samp", y="losses_samp", hue="method_samp", palette="tab10"
    )

    g = sns.lineplot(
        x=np.arange(
            max(df["batch_num_samp"]) + 1  # pylint: disable=unsubscriptable-object
        ),
        y=losses_cummin,
        color="black",
        ls="--",
        marker="o",
        label="min loss",
    )

    ids = df["method_samp"].unique()  # pylint: disable=unsubscriptable-object
    sampler_names = _get_samplers_names(saving_folder, ids)

    handles, labels = g.get_legend_handles_labels()
    labels = sampler_names + [labels[-1]]

    plt.legend(handles, labels, loc="upper right")


def plot_losses(saving_folder: Union[str, os.PathLike]) -> None:
    """Plot the parameter sampled colored according to their loss value.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))

    num_params = sum("params_samp_" in c_str for c_str in df.columns)

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        df,
        hue="losses_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="viridis",
        plot_kws=dict(markers="o", linewidth=1, alpha=0.8),
        diag_kws=dict(
            fill=False,
            hue=None,
        ),
    )

    g._legend.set_bbox_to_anchor((0.8, 0.5))  # pylint: disable=protected-access


def plot_sampling(saving_folder: Union[str, os.PathLike]) -> None:
    """Plot the parameter sampled colored according to the sampling method used to sample them.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))

    num_params = sum("params_samp_" in c_str for c_str in df.columns)
    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        df,
        hue="method_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="tab10",
        plot_kws=dict(markers="o", linewidth=1, alpha=0.8),
        diag_kws=dict(
            fill=False,
        ),
    )
    ids = df["method_samp"].unique()  # pylint: disable=unsubscriptable-object
    sampler_names = _get_samplers_names(saving_folder, ids)

    # take legend of the plot in the last row and first column, to be sure it's a scatter plot
    handles, _ = g.axes[-1][0].get_legend_handles_labels()
    g._legend.remove()  # pylint: disable=protected-access

    plt.legend(loc=2, handles=handles, labels=sampler_names, bbox_to_anchor=(0.0, 1.8))


def plot_losses_method_num(
    saving_folder: Union[str, os.PathLike], method_num: int
) -> None:
    """Plot the parameter sampled by a specific sampling method, and color them according to their loss value.

    Args:
        saving_folder: the folder where the calibration results were saved
        method_num: the integer value defining a specific sampling method
    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))

    if method_num not in set(df["method_samp"]):
        raise ValueError(f"Samplers with method_num = {method_num} was never used")

    df = df.loc[
        df["method_samp"] == method_num  # pylint: disable=unsubscriptable-object
    ]

    num_params = sum("params_samp_" in c_str for c_str in df.columns)

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        df,
        hue="losses_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="viridis",
        plot_kws=dict(markers="o", linewidth=1, alpha=0.8),
        diag_kws=dict(
            fill=False,
            hue=None,
        ),
    )

    g._legend.set_bbox_to_anchor((0.8, 0.5))  # pylint: disable=protected-access


def plot_losses_interact(saving_folder: Union[str, os.PathLike]) -> None:
    """Plot the parameter sampled colored according to their loss value.

    This plot allows to interactively choose the sampling method.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))
    method_nums = set(df["method_samp"])

    samplers_id_table = _get_samplers_id_table(saving_folder)

    id_samplers_cal = {}

    for sampler, sampler_id in samplers_id_table.items():
        if sampler_id in method_nums:
            id_samplers_cal[sampler] = sampler_id

    interact(
        plot_losses_method_num,
        saving_folder=fixed(saving_folder),
        method_num=id_samplers_cal,
    )


def plot_sampling_batch_nums(
    saving_folder: Union[str, os.PathLike], batch_nums: Collection[int]
) -> None:
    """Plot the parameter sampled in specific batches colored according to the sampling method used to sample them.

    Args:
        saving_folder: the folder where the calibration results were saved
        batch_nums: a list of batch number
    """
    # deduplicate batch numbers
    batch_nums = set(batch_nums)
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))

    filter_bns = [bn in batch_nums for bn in df["batch_num_samp"]]

    df["filter_bns"] = filter_bns

    df_ = df.loc[
        df["filter_bns"]  # pylint: disable=singleton-comparison,unsubscriptable-object
        == True  # noqa
    ]

    num_params = sum(
        "params_samp_" in c_str for c_str in df.columns  # pylint: disable=no-member
    )

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        df_,
        hue="method_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        plot_kws=dict(markers="o", linewidth=1, alpha=0.8),
        diag_kws=dict(fill=False),
        palette="tab10",
    )

    ids = df["method_samp"].unique()  # pylint: disable=unsubscriptable-object
    sampler_names = _get_samplers_names(saving_folder, ids)

    # take legend of the plot in the last row and first column, to be sure it's a scatter plot
    handles, _ = g.axes[-1][0].get_legend_handles_labels()
    g._legend.remove()  # pylint: disable=protected-access

    plt.legend(loc=2, handles=handles, labels=sampler_names, bbox_to_anchor=(0.0, 1.8))


def plot_sampling_interact(saving_folder: Union[str, os.PathLike]) -> None:
    """
    Plot the parameter sampled colored according to the sampling method used to sample them.

    The method allows to interactively choose the batch numbers included in the plot.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    df = pd.read_csv(os.path.join(saving_folder, "calibration_results.csv"))
    max_bn = int(max(df["batch_num_samp"]))
    all_bns = np.arange(max_bn + 1, dtype=int)
    indices_bns = np.array_split(all_bns, min(max_bn, 3))

    dict_bns = {}
    for index_bns in indices_bns:
        key = "from " + str(index_bns[0]) + " to " + str(index_bns[-1])
        dict_bns[key] = index_bns.tolist()

    interact(
        plot_sampling_batch_nums,
        saving_folder=fixed(saving_folder),
        batch_nums=dict_bns,
    )



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Method-of-moments loss function.

This module contains the implementation of the loss function
based on the 'method of moments'.
"""
from enum import Enum
from typing import Callable, List, Optional, Union, cast

import numpy as np
from numpy.typing import NDArray

from black_it.loss_functions.base import BaseLoss
from black_it.utils.base import is_symmetric
from black_it.utils.time_series import get_mom_ts_1d

MomentCalculator = Callable[[NDArray], NDArray]
"""
Type of a custom moment calculator: a function that takes an NDArray and returns
an NDArray.
"""


class _CovarianceMatrixType(Enum):
    """Enumeration of allowed covariance matrix types."""

    IDENTITY = "identity"
    INVERSE_VARIANCE = "inverse_variance"

    def __str__(self) -> str:
        """Get the string representation."""
        return self.value


class MethodOfMomentsLoss(BaseLoss):
    """Class for the 'method of moments' loss."""

    def __init__(
        self,
        covariance_mat: Union[str, NDArray[np.float64]] = "identity",
        coordinate_weights: Optional[NDArray[np.float64]] = None,
        moment_calculator: MomentCalculator = get_mom_ts_1d,
        coordinate_filters: Optional[List[Optional[Callable]]] = None,
        standardise_moments: bool = False,
    ):
        """
        Initialize the loss function based on the 'method of moments'.

        Returns the MSM objective function, i.e. the square difference between
        the moments of the two time series.
        By default the loss computes the moments using
        black_it.utils.time_series.get_mom_ts_1d(), which computes an
        18-dimensional vector of statistics.

        You can alter the behaviour passing a custom function to
        moment_calculator. Please note that there is a constraint between the
        moment calculator and the size of the covariance matrix.

        Args:
            covariance_mat: covariance matrix between the moments.
                'identity' (the default) gives the identity matrix, 'inverse_variance' gives the diagonal
                matrix of the estimated inverse variances. Alternatively, one can specify a numerical symmetric matrix
                whose size must be equal to the number of elements that the moment calculator returns.
            coordinate_weights: importance of each coordinate. By default all
                coordinates are treated equally.
            moment_calculator: a function that takes a 1D time series and
                returns a series of moments. The default is
                black_it.utils.time_series.get_mom_ts_1d()
            coordinate_filters: filters/transformations to be applied to each simulated series before
                the loss computation.
            standardise_moments: if True all moments are divided by the absolute value of the real moments,
                default is False.
        """
        MethodOfMomentsLoss._validate_covariance_and_calculator(
            moment_calculator, covariance_mat
        )

        super().__init__(coordinate_weights, coordinate_filters)
        self._covariance_mat = covariance_mat
        self._moment_calculator = moment_calculator
        self._standardise_moments = standardise_moments

    @staticmethod
    def _validate_covariance_and_calculator(
        moment_calculator: MomentCalculator,
        covariance_mat: Union[NDArray[np.float64], str],
    ) -> None:
        """
        Validate the covariance matrix.

        Args:
            moment_calculator: the moment calculator
            covariance_mat: the covariance matrix, either as a string or as a numpy array

        Returns:
            None

        Raises:
            ValueError: if the covariance matrix is not valid.
                If it is a string, ot can be invalid if it is not one of the possible options.
                If it is a numpy array, it can be invalid if it is not symmetric or if moment_calculator
                is the default get_mom_ts_1d and the covariance matrix's shape
                is not 18. Other possible errors won't be caught by this
                function, and can only be detected at runtime.
        """
        if isinstance(covariance_mat, str):

            try:
                _CovarianceMatrixType(covariance_mat)
            except ValueError as exc:
                raise ValueError(
                    f"expected one of {list(map(lambda x: x.value, _CovarianceMatrixType))}, got {covariance_mat})"
                ) from exc
            return

        if isinstance(covariance_mat, np.ndarray):
            # a non null covariance_mat was given
            if not is_symmetric(covariance_mat):
                raise ValueError(
                    "the provided covariance matrix is not valid as it is not a symmetric matrix"
                )
            if (moment_calculator is get_mom_ts_1d) and (covariance_mat.shape[0] != 18):
                raise ValueError(
                    "the provided covariance matrix is not valid as it has a wrong shape: "
                    f"expected 18, got {covariance_mat.shape[0]}"
                )
            return

        raise ValueError(
            "please specify a valid covariance matrix, either as a string or directly as a numpy array"
        )

    def compute_loss_1d(
        self, sim_data_ensemble: NDArray[np.float64], real_data: NDArray[np.float64]
    ) -> float:
        """
        Compute the loss based on the 'method of moments'.

        Returns the MSM objective function, i.e. the square difference between the moments of the two time series.

        Args:
            sim_data_ensemble: the first operand
            real_data: the second operand

        Returns:
            the MSM loss over a specific coordinate.
        """
        # compute the moments for the simulated ensemble
        ensemble_sim_mom_1d = np.array(
            [self._moment_calculator(s) for s in sim_data_ensemble]
        )

        # compute moments of the real time series
        real_mom_1d = self._moment_calculator(real_data)

        # write moments in relative terms is needed
        if self._standardise_moments:
            ensemble_sim_mom_1d = ensemble_sim_mom_1d / (abs(real_mom_1d)[None, :])
            real_mom_1d = real_mom_1d / abs(real_mom_1d)

        # mean simulated moments
        sim_mom_1d = np.mean(ensemble_sim_mom_1d, axis=0)

        g = real_mom_1d - sim_mom_1d

        if self._covariance_mat == _CovarianceMatrixType.IDENTITY.value:
            loss_1d = g.dot(g)
            return loss_1d
        if self._covariance_mat == _CovarianceMatrixType.INVERSE_VARIANCE.value:
            W = np.diag(
                1.0 / np.mean((real_mom_1d[None, :] - ensemble_sim_mom_1d) ** 2, axis=0)
            )
        else:
            self._covariance_mat = cast(NDArray[np.float64], self._covariance_mat)
            W = self._covariance_mat
        try:
            loss_1d = g.dot(W).dot(g)
        except ValueError as e:
            covariance_size = W.shape[0]
            moments_size = g.shape[0]

            if covariance_size == moments_size:
                # this value error is not due to a mismatch between the
                # covariance matrix size and the number moments. Let's raise the
                # original error.
                raise

            raise ValueError(
                f"The size of the covariance matrix ({covariance_size}) "
                f"and the number of moments ({moments_size}) should be identical"
            ) from e

        return loss_1d



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'MLSurrogateSampler' base class."""

from abc import abstractmethod
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.samplers.random_uniform import RandomUniformSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import digitize_data


class MLSurrogateSampler(BaseSampler):
    """
    MLSurrogateSampler interface.

    This is the base class for all machine learning surrogate samplers.
    """

    def __init__(
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_deduplication_passes: int = 5,
        candidate_pool_size: Optional[int] = None,
    ) -> None:
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the internal state of the sampler, fixing this numbers the sampler behaves deterministically
            max_deduplication_passes: maximum number of duplication passes done to avoid sampling repeated parameters
            candidate_pool_size: number of randomly sampled points on which the ML surrogate predictions are evaluated
        """
        super().__init__(batch_size, random_state, max_deduplication_passes)

        if candidate_pool_size is not None:
            self._candidate_pool_size = candidate_pool_size
        else:
            self._candidate_pool_size = 1000 * batch_size

    @property
    def candidate_pool_size(self) -> int:
        """Get the candidate pool size."""
        return self._candidate_pool_size

    def sample_candidates(
        self,
        candidate_pool_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Get a large pool of candidate parameters."""
        candidates = RandomUniformSampler(
            candidate_pool_size, random_state=self._get_random_seed()
        ).sample_batch(
            candidate_pool_size, search_space, existing_points, existing_losses
        )
        return candidates

    @abstractmethod
    def fit(self, X: NDArray[np.float64], y: NDArray[np.float64]) -> None:
        """Abstract method to fit the loss function of an ML surrogate."""

    @abstractmethod
    def predict(self, X: NDArray[np.float64]) -> NDArray[np.float64]:
        """Abstract method for the predictions of an ML surrogate."""

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample a number of new parameters fixed by the 'batch_size' attribute.

        Args:
            batch_size: number of samples to collect
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the new parameters
        """
        # Get a large pool of potential candidates
        candidates = self.sample_candidates(
            self.candidate_pool_size, search_space, existing_points, existing_losses
        )

        # Train surrogate model
        self.fit(existing_points, existing_losses)

        # Predictions of surrogate on large pool of candidates
        predictions = self.predict(candidates)

        # Select candidates with lowest predicted loss value
        sorting_indices: NDArray[np.int64] = np.argsort(predictions)
        sampled_points: NDArray[np.float64] = candidates[sorting_indices][:batch_size]

        return digitize_data(sampled_points, search_space.param_grid)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the XGBoost sampling."""
import warnings
from typing import Optional, cast

import numpy as np
import xgboost as xgb
from numpy.typing import NDArray

from black_it.samplers.surrogate import MLSurrogateSampler

MAX_FLOAT32 = np.finfo(np.float32).max
MIN_FLOAT32 = np.finfo(np.float32).min
EPS_FLOAT32 = np.finfo(np.float32).eps


class XGBoostSampler(MLSurrogateSampler):
    """This class implements xgboost sampling."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_deduplication_passes: int = 5,
        candidate_pool_size: Optional[int] = None,
        colsample_bytree: float = 0.3,
        learning_rate: float = 0.1,
        max_depth: int = 5,
        alpha: float = 1.0,
        n_estimators: int = 10,
    ) -> None:
        """
        Sampler based on a xgboost surrogate model of the loss function.

        Note: this class makes use of the xgboost library, for more information on the XGBoost parameters
            visit https://xgboost.readthedocs.io/en/stable/parameter.html.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: the maximum number of deduplication passes
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            colsample_bytree: subsample ratio of columns when constructing each tree
            learning_rate: the learning rate
            max_depth: maximum depth of XGBoost trees
            alpha: L1 regularization term on weights
            n_estimators: number of estimators

        References:
            Lamperti, Roventini, and Sani, "Agent-based model calibration using machine learning surrogates"
        """
        super().__init__(
            batch_size, random_state, max_deduplication_passes, candidate_pool_size
        )

        self._colsample_bytree = colsample_bytree
        self._learning_rate = learning_rate
        self._max_depth = max_depth
        self._alpha = alpha
        self._n_estimators = n_estimators
        self._xg_regressor: Optional[xgb.XGBRegressor] = None

    @property
    def colsample_bytree(self) -> float:
        """Get the colsample_bytree parameter."""
        return self._colsample_bytree

    @property
    def learning_rate(self) -> float:
        """Get the learning rate."""
        return self._learning_rate

    @property
    def max_depth(self) -> int:
        """Get the maximum tree depth."""
        return self._max_depth

    @property
    def alpha(self) -> float:
        """Get the alpha regularisation parameter."""
        return self._alpha

    @property
    def n_estimators(self) -> int:
        """Get the number of estimators."""
        return self._n_estimators

    @staticmethod
    def _clip_losses(y: NDArray[np.float64]) -> NDArray[np.float64]:
        """Check that loss values fall within the float32 limits needed for XGBoost to work."""
        large_floats = np.where(y >= MAX_FLOAT32)
        small_floats = np.where(y <= MIN_FLOAT32)

        if len(large_floats) == 0 and len(small_floats) == 0:
            return y

        warnings.warn(
            "Found loss values out of float32 limits, clipping them for XGBoost.",
            RuntimeWarning,
        )
        if len(large_floats) > 0:
            y[large_floats] = MAX_FLOAT32 - EPS_FLOAT32

        if len(small_floats) > 0:
            y[small_floats] = MIN_FLOAT32 + EPS_FLOAT32

        return y

    def fit(self, X: NDArray[np.float64], y: NDArray[np.float64]) -> None:
        """Fit a xgboost surrogate model."""
        # prepare data
        y = self._clip_losses(y)  # pylint: disable=W0212
        _ = xgb.DMatrix(data=X, label=y)

        # train surrogate
        self._xg_regressor = xgb.XGBRegressor(
            objective="reg:squarederror",  # original: objective ='reg:linear',
            colsample_bytree=self.colsample_bytree,
            learning_rate=self.learning_rate,
            max_depth=self.max_depth,  # original: 5
            alpha=self.alpha,
            n_estimators=self.n_estimators,
        )  # original: 10

        self._xg_regressor.fit(X, y)

    def predict(self, X: NDArray[np.float64]) -> NDArray[np.float64]:
        """Predict using a xgboost surrogate model."""
        # predict over large pool of candidates
        _ = xgb.DMatrix(data=X)

        self._xg_regressor = cast(xgb.XGBRegressor, self._xg_regressor)
        predictions = self._xg_regressor.predict(X)

        return predictions
